//-----------------------------------------------------------------------
// class CNetwork
//
// Network core
//
// Author:  Sylvain Lefebvre
// History: Created 02/04/2002
//-----------------------------------------------------------------------
#ifndef __CNETWORK__
#define __CNETWORK__
//-----------------------------------------------------------------------
#define CNETWORK_VER_MAJ 0
#define CNETWORK_VER_MIN 5
//-----------------------------------------------------------------------
#include "message_types.h"
//-----------------------------------------------------------------------
#include <list>
#include <map>
using namespace std;
//-----------------------------------------------------------------------

#include <SDL/SDL.h>
#include <SDL/SDL_net.h>
#include <SDL/SDL_thread.h>
//-----------------------------------------------------------------------
typedef struct network_data
{
  unsigned char *m_Raw;
  int            m_iSize;
}network_data;
//-----------------------------------------------------------------------
typedef struct network_client
{
  TCPsocket m_Socket;
  IPaddress m_UDPAddress;
  Uint8     m_Id;
}network_client;
//-----------------------------------------------------------------------
typedef struct network_msg
{
  Uint32          m_Type;
  Uint32          m_Util;
  Uint8           m_ClientId;
  network_data    m_Data;
  IPaddress       m_Ip;
  TCPsocket       m_Socket;
  bool            m_bFromServer;
  bool            m_bLocal;
}network_msg;
//-----------------------------------------------------------------------
#define CNETWORK_UDPPACKET_SIZE 1024
#define CNETWORK_ID             "WOD1.0"
#define CNETWORK_REFUSE_MSG     "Refusing connection: bad id, should be "
#define CNETWORK_MUTEX_TMOUT    1000
#define CNETWORK_SERVER_ID      254
#define CNETWORK_UNKNOWN_ID     255
//-----------------------------------------------------------------------
#define NTWRK_FAST              1
#define NTWRK_SECURE            2
#define NTWRK_TO_SERVER         256
#define NTWRK_TO_OTHERS         257
#define NTWRK_TO_ALL            258
//-----------------------------------------------------------------------
#define NETWORK (*CNetwork::getUniqueInstance())
#define NTWRK   NETWORK
//-----------------------------------------------------------------------
class INetworkUser;
//-----------------------------------------------------------------------
class CNetwork
{
private:
  
  class ip_equal
  {
  public:
    bool operator () (const IPaddress& a0,const IPaddress& a1)
      {
	if (a0.host < a1.host)
	  return (true);
	else if (a0.host == a1.host)
	  return (a0.port < a1.port);
	else
	  return (false);
      }
  };

  map<int,network_client> m_Clients;
  typedef map<int,network_client>::iterator client_iterator;

  static Uint16        s_iPort;
  static Uint16        s_iClientUDPPort;

  INetworkUser        *m_User;
  list<network_msg>    m_Fifo;
  bool                 m_bServerStarted;
  bool                 m_bClientStarted;
  bool                 m_bNetworkAvailable;
  bool                 m_bThreadShouldStop;
  bool                 m_bReady;
  bool                 m_bError;
  IPaddress            m_IpServer;

  TCPsocket            m_TCPSocket;
  UDPsocket            m_UDPSocket;
    
  SDL_Thread          *m_TCPThread;
  SDL_Thread          *m_UDPThread;

  static SDL_mutex    *s_MutexFifo;
  static SDL_mutex    *s_MutexBuffer;

  unsigned char       *m_Buffer;
  int                  m_iLenBuffer;
  network_msg          m_TmpMsg;

  int                  m_iClientId; 
  int                  m_iNextClientId; 
  unsigned char        m_ui8VerMaj;
  unsigned char        m_ui8VerMin;
  unsigned char        m_ui8VerRev;

  friend int CNetwork_TCP_server_thread(void *);           
  friend int CNetwork_TCP_client_thread(void *);           
  friend int CNetwork_UDP_thread(void *);

  bool       threadShouldStop() {return (m_bThreadShouldStop);}

  static int getTCPData(TCPsocket,unsigned char **);
  static int sendTCPData(TCPsocket,unsigned char *,int);
  static int getUDPData(UDPsocket,unsigned char **);
  static int sendUDPData(UDPsocket,const IPaddress&,unsigned char *,int);

  inline static void lockFifo();
  inline static void unlockFifo();
  inline static void lockBuffer();
  inline static void unlockBuffer();

  void startServerThreads();
  void startClientThreads();

  static void  allocBuffer(int,unsigned char *&,int&);
  int          prepareMessage(network_msg *,unsigned char *&,int&);
  void         extractMessage(network_msg&,unsigned char *,int);
  void         extractMessage(const TCPsocket,const IPaddress *,
			      unsigned char *,int);
  void         storeMessage(const network_msg *);
  void         freeMessageData(network_msg&);
  network_msg *buildRelayMessage(network_msg *src,int,network_msg *dst);
  network_msg *extractRelayMessage(network_msg *src,network_msg *dst);
  
  void         updateNetwork(const network_client&);

  void         secureSend(TCPsocket,network_msg *);
  void         fastSend(const IPaddress&,network_msg *);

  network_msg *messageFromData(int,Uint8,unsigned char*,int,network_msg *);

protected:

  virtual void onClientConnect(const TCPsocket&);
  virtual void onClientDisconnect(const TCPsocket&);

  CNetwork();
  virtual ~CNetwork();

public:

  static CNetwork *getUniqueInstance();

  void startServer(unsigned char, unsigned char, unsigned char);
  void startClient(char *, unsigned char, unsigned char, unsigned char);

  network_msg *clientMessageFromData(int,unsigned char*,int,network_msg *);
  network_msg *serverMessageFromData(int,unsigned char*,int,network_msg *);

  void         serverSend(int method,int dest,network_msg *,
			  bool relay=false);
  void         clientSend(int method,int dest,network_msg *);

  void         processMessages();
  
  bool         isServer();
  bool         isClient();
  bool         isEnabled();
  bool         isReady();

  void         setClientUDPPort(Uint16 p) {s_iClientUDPPort=p;}
  int          clientId() {return (m_iClientId);}

  void         bindTo(INetworkUser *);

  void         print(const char *msg,...);
  void         printLog(const char *msg,...);

  bool         error() const {return (m_bError);}
};
//-----------------------------------------------------------------------
#endif
//-----------------------------------------------------------------------
// ---------------------------------------------------------- 
// Sylvain Lefebvre - (c) 2002 - Sylvain.Lefebvre@laposte.net
// ----------------------------------------------------------
